{
    'Analysis': {
        'required': False,
        'type': 'list',
        'nullable': True,
        'schema': {
            'type': 'dict',
            'allow_unknown': False,
            'oneof_schema': [
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['AllowlistViolationDetector'], 'required': True},
                    'allowlist_rules': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['MatchPathFilter'], 'required': True},
                    'parsed_atom_handler_lookup_list': {
                        'type': 'list', 'schema': {'type': 'list', 'schema': {'type': 'string', 'nullable': True}}, 'required': True},
                    'default_parsed_atom_handler': {'type': 'string', 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['MatchValueFilter'], 'required': True},
                    'path': {'type': 'string', 'required': True, 'empty': False},
                    'parsed_atom_handler_dict': {
                        'type': 'dict', 'schema': {'id': {'type': 'string'}, 'type': {'type': 'string'}}, 'required': True},
                    'default_parsed_atom_handler': {'type': 'string', 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True},
                    'type': {'type': 'string', 'allowed': ['PCADetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string'}, 'nullable': True},
                    'window_size': {'type': ['integer', 'float'], 'min': 0.001},
                    'min_anomaly_score': {'type': 'float'},
                    'min_variance': {'type': 'float'},
                    'num_windows': {'type': 'float'},
                    'persistence_id': {'type': 'string'},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'constraint_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'nullable': True}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['EnhancedNewMatchPathValueComboDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True, 'required': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'allow_missing_values': {'type': 'boolean'},
                    'learn_mode': {'type': 'boolean'},
                    'tuple_transformation_function': {'type': 'string', 'allowed': ['demo'], 'nullable': True},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['EventCorrelationDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'max_hypotheses': {'type': 'integer', 'min': 1},
                    'hypothesis_max_delta_time': {'type': 'float', 'min': 0.01},
                    'generation_probability': {'type': 'float', 'min': 0, 'max': 1},
                    'generation_factor': {'type': 'float', 'min': 0, 'max': 1},
                    'max_observations': {'type': 'integer', 'min': 1},
                    'p0': {'type': 'float', 'min': 0, 'max': 1},
                    'alpha': {'type': 'float', 'min': 0, 'max': 1},
                    'candidates_size': {'type': 'integer', 'min': 1},
                    'hypotheses_eval_delta_time': {'type': 'float', 'min': 0.01},
                    'delta_time_to_discard_hypothesis': {
                        'type': 'float', 'min': 0.01, 'bigger_than_or_equal': ['hypotheses_eval_delta_time', 120.0]},
                    'check_rules_flag': {'type': 'boolean'},
                    'learn_mode': {'type': 'boolean'},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'output_logline': {'type': 'boolean'},
                    'constraint_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['EventFrequencyDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'scoring_path_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'unique_path_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'window_size': {'type': ['integer', 'float'], 'min': 0.001},
                    'num_windows': {'type': 'integer'},
                    'empty_window_warnings': {'type': 'boolean'},
                    'early_exceeding_anomaly_output': {'type': 'boolean'},
                    'set_lower_limit': {'type': 'integer', 'min': 0, 'nullable': True},
                    'set_upper_limit': {'type': 'integer', 'min': 0, 'nullable': True},
                    'confidence_factor': {'type': 'float', 'min': 0, 'max': 1},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'constraint_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['EventCountClusterDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'id_path_list': {'type': 'list', 'nullable': True},
                    'window_size': {'type': ['integer', 'float'], 'min': 0.001},
                    'num_windows': {'type': 'integer'},
                    'confidence_factor': {'type': 'float', 'min': 0, 'max': 1},
                    'idf': {'type': 'boolean'},
                    'norm': {'type': 'boolean'},
                    'add_normal': {'type': 'boolean'},
                    'check_empty_windows': {'type': 'boolean'},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'constraint_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['EventSequenceDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'id_path_list': {'type': 'list', 'nullable': True},
                    'seq_len': {'type': 'integer', 'min': 1},
                    'timeout': {'type': 'integer', 'min': -1},
                    'allow_missing_id': {'type': 'boolean'},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'constraint_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['ValueRangeDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'id_path_list': {'type': 'list', 'nullable': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'constraint_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['CharsetDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'id_path_list': {'type': 'list', 'nullable': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'constraint_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['EntropyDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'prob_thresh': {'type': 'float'},
                    'default_freqs': {'type': 'boolean'},
                    'skip_repetitions': {'type': 'boolean'},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'constraint_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['EventTypeDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'id_path_list': {'type': 'list', 'nullable': True},
                    'allow_missing_id': {'type': 'boolean'},
                    'allowed_id_tuples': {'type': 'list', 'schema': {'type': 'list', 'schema': {'type': 'string'}}, 'nullable': True},
                    'min_num_vals': {'type': 'integer', 'min': 1},
                    'max_num_vals': {'type': 'integer', 'min': 1, 'bigger_than_or_equal': ['min_num_vals', 1000]},
                    'save_values': {'type': 'boolean'},
                    'learn_mode': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['SlidingEventFrequencyDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'scoring_path_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'window_size': {'type': ['integer', 'float'], 'min': 0.001},
                    'set_upper_limit': {'type': 'integer', 'min': 0, 'nullable': True},
                    'local_maximum_threshold': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'constraint_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['LinearNumericBinDefinition'], 'required': True},
                    'lower_limit': {'type': ['integer', 'float'], 'required': True},
                    'bin_size': {'type': 'integer', 'required': True, 'min': 1},
                    'bin_count': {'type': 'integer', 'required': True, 'min': 1},
                    'outlier_bins_flag': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['ModuloTimeBinDefinition'], 'required': True},
                    'modulo_value': {'type': ['integer', 'float'], 'required': True, 'min': 0.000001},
                    'time_unit': {'type': 'integer', 'required': True, 'min': 1},
                    'lower_limit': {'type': ['integer', 'float'], 'required': True, 'min': 0},
                    'bin_size': {'type': 'integer', 'required': True, 'min': 1},
                    'bin_count': {'type': 'integer', 'required': True, 'min': 1},
                    'outlier_bins_flag': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['HistogramAnalysis'], 'required': True},
                    'histogram_defs': {
                        'type': 'list', 'schema': {'type': 'list', 'schema': {'type': 'string', 'empty': False}}, 'required': True},
                    'report_interval': {'type': 'integer', 'required': True, 'min': 1},
                    'reset_after_report_flag': {'type': 'boolean'},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['PathDependentHistogramAnalysis'], 'required': True},
                    'path': {'type': 'string', 'required': True, 'empty': False},
                    'bin_definition': {'type': 'string', 'required': True, 'empty': False},
                    'report_interval': {'type': 'integer', 'required': True, 'min': 1},
                    'reset_after_report_flag': {'type': 'boolean'},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['MatchFilter'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'value_list': {
                        'type': 'list', 'schema': {'type': ['boolean', 'float', 'integer', 'string']}, 'nullable': True},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['MatchValueAverageChangeDetector'], 'required': True},
                    'timestamp_path': {'type': 'string', 'required': True, 'nullable': True, 'empty': False},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'min_bin_elements': {'type': 'integer', 'required': True, 'min': 1},
                    'min_bin_time': {'type': 'integer', 'required': True, 'min': 1},
                    'debug_mode': {'type': 'boolean'},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['MatchValueStreamWriter'], 'required': True},
                    'stream': {'type': 'string', 'allowed': ['sys.stdout', 'sys.stderr'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'separator': {'type': 'string', 'required': True},
                    'missing_value_string': {'type': 'string', 'required': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['MissingMatchPathValueDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'learn_mode': {'type': 'boolean'},
                    'combine_values': {'type': 'boolean'},
                    'check_interval': {'type': 'integer', 'min': 1},
                    'realert_interval': {'type': 'integer', 'min': 1, 'bigger_than_or_equal': ['check_interval', 3600]},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['MissingMatchPathListValueDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'learn_mode': {'type': 'boolean'},
                    'combine_values': {'type': 'boolean'},
                    'check_interval': {'type': 'integer', 'min': 1},
                    'realert_interval': {'type': 'integer', 'min': 1, 'bigger_than_or_equal': ['check_interval', 3600]},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['NewMatchIdValueComboDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'id_path_list': {'type': 'list', 'required': True},
                    'min_allowed_time_diff': {'type': 'float', 'required': True, 'min': 0.01},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'allow_missing_values': {'type': 'boolean'},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['NewMatchPathDetector'], 'required': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['NewMatchPathValueComboDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'allow_missing_values': {'type': 'boolean'},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['NewMatchPathValueDetector'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'learn_mode': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['ParserCount'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}},
                    'report_interval': {'type': 'integer', 'min': 1},
                    'labels': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'split_reports_flag': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['EventGenerationMatchAction'], 'required': True},
                    'event_type': {'type': 'string', 'required': True},
                    'event_message': {'type': 'string', 'required': True}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['AtomFilterMatchAction'], 'required': True},
                    # this is optional on purpose. If not used, the default atom_filter is used.
                    'subhandler_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}},
                    'stop_when_handled_flag': {'type': 'boolean'},
                    'delete_components': {'type': 'boolean'}
                    },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['AndMatchRule', 'OrMatchRule', 'ParallelMatchRule'], 'required': True},
                    'sub_rules': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'match_action': {'type': 'string', 'nullable': True}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['ValueDependentDelegatedMatchRule'], 'required': True},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'rule_lookup_dict': {'type': 'dict', 'valuesrules': {'type': 'string'}, 'required': True},
                    'default_rule': {'type': 'string', 'nullable': True, 'empty': False},
                    'match_action': {'type': 'string', 'nullable': True, 'empty': False}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['NegationMatchRule'], 'required': True},
                    'sub_rule': {'type': 'string', 'required': True, 'empty': False},
                    'match_action': {'type': 'string', 'nullable': True, 'empty': False}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['PathExistsMatchRule', 'IPv4InRFC1918MatchRule'], 'required': True},
                    'path': {'type': 'string', 'required': True, 'empty': False},
                    'match_action': {'type': 'string', 'nullable': True, 'empty': False}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['ValueMatchRule'], 'required': True},
                    'path': {'type': 'string', 'required': True, 'empty': False},
                    'value': {'type': ['boolean', 'float', 'integer', 'string'], 'required': True},
                    'match_action': {'type': 'string', 'nullable': True, 'empty': False}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['ValueListMatchRule'], 'required': True},
                    'path': {'type': 'string', 'required': True, 'empty': False},
                    'value_list': {'type': 'list', 'schema': {'type': ['boolean', 'float', 'integer', 'string']}, 'required': True},
                    'match_action': {'type': 'string', 'nullable': True, 'empty': False}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['ValueRangeMatchRule'], 'required': True},
                    'path': {'type': 'string', 'required': True, 'empty': False},
                    'lower_limit': {'type': ['integer', 'float'], 'required': True},
                    'upper_limit': {'type': ['integer', 'float'], 'required': True, 'bigger_than_or_equal': ['lower_limit', None]},
                    'match_action': {'type': 'string', 'nullable': True, 'empty': False}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['StringRegexMatchRule'], 'required': True},
                    'path': {'type': 'string', 'required': True, 'empty': False},
                    'regex': {'type': 'string', 'required': True, 'empty': False},
                    'match_action': {'type': 'string', 'nullable': True, 'empty': False}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['ModuloTimeMatchRule'], 'required': True},
                    'path': {'type': 'string', 'required': True, 'nullable': True, 'empty': False},
                    'seconds_modulo': {'type': 'integer', 'required': True, 'min': 1},
                    'lower_limit': {'type': ['integer', 'float'], 'required': True, 'min': 0},
                    'upper_limit': {
                        'type': ['integer', 'float'], 'required': True, 'min': 0, 'bigger_than_or_equal': ['lower_limit', None]},
                    'match_action': {'type': 'string', 'nullable': True, 'empty': False}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['ValueDependentModuloTimeMatchRule'], 'required': True},
                    'path': {'type': 'string', 'required': True, 'nullable': True, 'empty': False},
                    'seconds_modulo': {'type': 'integer', 'required': True, 'min': 1},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'limit_lookup_dict': {'type': 'dict', 'valuesrules': {'type': 'list', 'schema': {
                            'type': ['integer', 'float'], 'min': 0}}, 'required': True},
                    'default_limit': {'type': 'list', 'schema': {'type': 'integer', 'min': 0}, 'nullable': True},
                    'match_action': {'type': 'string', 'nullable': True, 'empty': False}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['DebugMatchRule', 'DebugHistoryMatchRule'], 'required': True},
                    'debug_mode': {'type': 'boolean'},
                    'match_action': {'type': 'string', 'nullable': True, 'empty': False}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['TimeCorrelationDetector'], 'required': True},
                    'parallel_check_count': {'type': 'integer', 'required': True, 'min': 1},
                    'record_count_before_event': {'type': 'integer', 'min': 1},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'output_logline': {'type': 'boolean'},
                    'use_path_match': {'type': 'boolean'},
                    'use_value_match': {'type': 'boolean'},
                    'min_rule_attributes': {'type': 'integer', 'min': 1},
                    'max_rule_attributes': {'type': 'integer', 'min': 1, 'bigger_than_or_equal': ['min_rule_attributes', 1]},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['TimeCorrelationViolationDetector'], 'required': True},
                    'ruleset': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'required': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'type': {'type': 'string', 'allowed': ['CorrelationRule'], 'required': True},
                    'rule_id': {'type': 'string', 'required': True, 'empty': False},
                    'min_time_delta': {'type': 'integer', 'required': True, 'min': 1},
                    'max_time_delta': {'type': 'integer', 'required': True, 'min': 1, 'bigger_than_or_equal': ['min_time_delta', None]},
                    'artefact_match_parameters': {'type': 'list', 'schema': {'type': 'list', 'schema': {'type': 'string', 'empty': False}},
                                                  'nullable': True}
                },
                {
                    'type': {'type': 'string', 'allowed': ['EventClassSelector'], 'required': True},
                    'action_id': {'type': 'string', 'required': True, 'empty': False},
                    'artefact_a_rules': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'artefact_b_rules': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['TimestampsUnsortedDetector'], 'required': True},
                    'exit_on_error_flag': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['VariableCorrelationDetector'], 'required': True},
                    'event_type_detector': {'type': 'string', 'required': True, 'empty': False},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'num_init': {'type': 'integer', 'min': 1},
                    'num_update': {'type': 'integer', 'min': 1},
                    'disc_div_thres': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'num_steps_create_new_rules': {'type': 'integer', 'min': 1},
                    'num_upd_until_validation': {'type': 'integer', 'min': 1},
                    'num_end_learning_phase': {'type': 'integer', 'min': 1},
                    'check_cor_thres': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'check_cor_prob_thres': {'type': 'float', 'min': 0.000001},
                    'check_cor_num_thres': {'type': 'integer', 'min': 1},
                    'min_values_cors_thres': {'type': 'integer', 'min': 1},
                    'new_vals_alarm_thres': {'type': 'float', 'min': 0.000001},
                    'num_bt': {'type': 'integer', 'min': 1},
                    'alpha_bt': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'used_homogeneity_test': {'type': 'string', 'allowed': ['Chi', 'MaxDist']},
                    'alpha_chisquare_test': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'max_dist_rule_distr': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'used_presel_meth': {'type': 'list', 'schema': {'type': 'string', 'allowed': [
                        'matchDiscDistr', 'excludeDueDistr', 'matchDiscVals', 'random']}, 'nullable': True},
                    'intersect_presel_meth': {'type': 'boolean'},
                    'percentage_random_cors': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'match_disc_vals_sim_tresh': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'exclude_due_distr_lower_limit': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'match_disc_distr_threshold': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'used_cor_meth': {'type': 'list', 'schema': {'type': 'string', 'allowed': ['Rel', 'WRel']}, 'nullable': True},
                    'used_validate_cor_meth': {'type': 'list', 'schema': {'type': 'string', 'allowed': [
                        'coverVals', 'distinctDistr']}, 'nullable': True},
                    'validate_cor_cover_vals_thres': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'validate_cor_distinct_thres': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'constraint_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['VariableTypeDetector'], 'required': True},
                    'event_type_detector': {'type': 'string', 'required': True, 'empty': False},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'used_gof_test': {'type': 'string', 'allowed': ['CM', 'KS']},
                    'gof_alpha': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    's_gof_alpha': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    's_gof_bt_alpha': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'd_alpha': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'd_bt_alpha': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'div_thres': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'sim_thres': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'indicator_thres': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'num_init': {'type': 'integer', 'min': 1},
                    'num_update': {'type': 'integer', 'min': 1},
                    'num_update_unq': {'type': 'integer', 'min': 1},
                    'num_s_gof_values': {'type': 'integer', 'min': 1},
                    'num_s_gof_bt': {'type': 'integer', 'min': 1},
                    'num_d_bt': {'type': 'integer', 'min': 1},
                    'num_pause_discrete': {'type': 'integer', 'min': 0},
                    'num_pause_others': {'type': 'integer', 'min': 0},
                    'test_gof_int': {'type': 'boolean'},
                    'num_stop_update': {'type': 'boolean'},
                    'silence_output_without_confidence': {'type': 'boolean'},
                    'silence_output_except_indicator': {'type': 'boolean'},
                    'num_var_type_hist_ref': {'type': 'integer', 'min': 1},
                    'num_update_var_type_hist_ref': {'type': 'integer', 'min': 1},
                    'num_var_type_considered_ind': {'type': 'integer', 'min': 1},
                    'num_stat_stop_update': {'type': 'integer', 'min': 1},
                    'num_updates_until_var_reduction': {'type': 'integer', 'min': 0},
                    'var_reduction_thres': {'type': 'float'},
                    'num_skipped_ind_for_weights': {'type': 'integer', 'min': 0},
                    'num_ind_for_weights': {'type': 'integer', 'min': 1},
                    'used_multinomial_test': {'type': 'string', 'allowed': ['Approx', 'MT', 'Chi']},
                    'use_empiric_distr': {'type': 'boolean'},
                    'range_alpha': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'used_range_test': {'type': 'string', 'allowed': ['MeanSD', 'EmpiricQuantiles', 'MinMax']},
                    'range_threshold': {'type': 'float', 'min': 0.000001},
                    'range_limits_factor': {'type': 'float', 'min': 0.000001},
                    'num_reinit_range': {'type': 'integer', 'min': 0},
                    'dw_alpha': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'save_statistics': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'constraint_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'learn_mode': {'type': 'boolean'},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['PathValueTimeIntervalDetector'], 'required': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'time_period_length': {'type': 'integer', 'min': 1},
                    'max_time_diff': {'type': 'integer', 'min': 1},
                    'num_reduce_time_list': {'type': 'integer', 'min': 1},
                    'allow_missing_values': {'type': 'boolean'},
                    'output_logline': {'type': 'boolean'},
                    'learn_mode': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['PathArimaDetector'], 'required': True},
                    'event_type_detector': {'type': 'string', 'required': True, 'empty': False},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_logline': {'type': 'boolean'},
                    'num_init': {'type': 'integer', 'min': 1},
                    'force_period_length': {'type': 'boolean'},
                    'set_period_length': {'type': 'integer', 'min': 1},
                    'alpha': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'alpha_bt': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'num_results_bt': {'type': 'integer', 'min': 1},
                    'num_min_time_history': {'type': 'integer', 'min': 1},
                    'num_max_time_history': {'type': 'integer', 'min': 2},
                    'num_periods_tsa_ini': {'type': 'integer', 'min': 2},
                    'learn_mode': {'type': 'boolean'},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['TSAArimaDetector'], 'required': True},
                    'event_type_detector': {'type': 'string', 'required': True, 'empty': False},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'waiting_time': {'type': 'integer', 'min': 1},
                    'num_sections_waiting_time': {'type': 'integer', 'min': 1},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'output_logline': {'type': 'boolean'},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'acf_pause_interval_percentage': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'acf_auto_pause_interval': {'type': 'boolean'},
                    'acf_auto_pause_interval_num_min': {'type': 'integer', 'min': 1},
                    'build_sum_over_values': {'type': 'boolean'},
                    'num_periods_tsa_ini': {'type': 'integer', 'min': 2},
                    'num_division_time_step': {'type': 'integer', 'min': 1},
                    'alpha': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'num_min_time_history': {'type': 'integer', 'min': 1},
                    'num_max_time_history': {'type': 'integer', 'min': 2},
                    'num_results_bt': {'type': 'integer', 'min': 1},
                    'alpha_bt': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'round_time_interval_threshold': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'acf_threshold': {'type': 'float', 'min': 0.000001, 'max': 1.0},
                    'force_period_length': {'type': 'boolean'},
                    'set_period_length': {'type': 'integer', 'min': 1},
                    'min_log_lines_per_time_step': {'type': 'integer', 'min': 1},
                    'output_event_handlers': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'learn_mode': {'type': 'boolean'},
                    'suppress': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['MinimalTransitionTimeDetector'], 'required': True},
                    'persistence_id': {'type': 'string', 'empty': False},
                    'paths': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'id_path_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'ignore_list': {'type': 'list', 'schema': {'type': 'string', 'empty': False}, 'nullable': True},
                    'allow_missing_id': {'type': 'boolean'},
                    'num_log_lines_solidify_matrix': {'type': 'integer', 'min': 1},
                    'time_output_threshold': {'type': 'integer', 'min': 0},
                    'anomaly_threshold': {'type': 'float', 'min': 0, 'max': 1.0},
                    'output_logline': {'type': 'boolean'},
                    'learn_mode': {'type': 'boolean'}
                },
                {
                    'id': {'type': 'string', 'nullable': True, 'empty': False},
                    'type': {'type': 'string', 'allowed': ['SimpleUnparsedAtomHandler', 'VerboseUnparsedAtomHandler'], 'required': True},
                    'suppress': {'type': 'boolean'}
                }
            ]
        }
    }
}



import unittest
from aminer.analysis.NewMatchPathDetector import NewMatchPathDetector
from aminer.analysis.NewMatchPathValueComboDetector import NewMatchPathValueComboDetector
from aminer.input.LogAtom import LogAtom
import time
from datetime import datetime
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.ParserMatch import ParserMatch
from unit.TestBase import TestBase


class NewMatchPathDetectorTest(TestBase):
    """Unittests for the NewMatchPathDetector."""

    __expected_string = '%s New path(es) detected\n%s: "%s" (%d lines)\n  %s\n%s\n\n'
    match_path_s1 = "['/s1']"
    match_path_d1 = "['/d1']"

    datetime_format_string = '%Y-%m-%d %H:%M:%S'
    analysis = 'Analysis.%s'
    pid = " pid="
    uid = " uid=2"

    match_context_fixed_dme = MatchContext(b' pid=')
    fixed_dme = FixedDataModelElement('s1', b' pid=')
    match_element_fixed_dme = fixed_dme.get_match_element("", match_context_fixed_dme)

    match_context_decimal_integer_value_me = MatchContext(b'25537 uid=2')
    decimal_integer_value_me = DecimalIntegerValueModelElement('d1', DecimalIntegerValueModelElement.SIGN_TYPE_NONE,
                                                               DecimalIntegerValueModelElement.PAD_TYPE_NONE)
    match_element_decimal_integer_value_me = decimal_integer_value_me.get_match_element("", match_context_decimal_integer_value_me)

    def test1_log_atom_not_known(self):
        """
        This test case checks the correct processing of unknown log lines, which in reality means that an anomaly has been found.
        The output is directed to an output stream and compared for accuracy. The learn_mode is False and the output must be
        repeatable on second run.
        """
        description = "Test1NewMatchPathDetector"
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', False,
                                                       output_logline=False)
        self.analysis_context.register_component(new_match_path_detector, description)
        t = round(time.time(), 3)
        log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_fixed_dme), t, new_match_path_detector)
        log_atom_decimal_integer_value_me = LogAtom(self.match_context_decimal_integer_value_me.match_data,
                                                    ParserMatch(self.match_element_decimal_integer_value_me), t, new_match_path_detector)

        self.assertTrue(new_match_path_detector.receive_atom(log_atom_fixed_dme))
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string % (
            datetime.fromtimestamp(t).strftime(self.datetime_format_string), new_match_path_detector.__class__.__name__, description, 1,
            self.match_path_s1, self.pid))
        self.reset_output_stream()

        # repeating should produce the same result
        self.assertTrue(new_match_path_detector.receive_atom(log_atom_fixed_dme))
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string % (
            datetime.fromtimestamp(t).strftime(self.datetime_format_string), new_match_path_detector.__class__.__name__, description, 1,
            self.match_path_s1, self.pid))
        self.reset_output_stream()

        # other MatchElement
        self.assertTrue(new_match_path_detector.receive_atom(log_atom_decimal_integer_value_me))
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string % (
            datetime.fromtimestamp(t).strftime(self.datetime_format_string), new_match_path_detector.__class__.__name__, description, 1,
            self.match_path_d1, self.uid))

    def test2_log_atom_known(self):
        """
        This test case checks the functionality of the learn_mode.
        If the same MatchElement is processed a second time and the learn_mode was True, no event must be triggered.
        """
        description = "Test2NewMatchPathDetector"
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True,
                                                       output_logline=False)
        self.analysis_context.register_component(new_match_path_detector, description)
        t = round(time.time(), 3)

        log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_fixed_dme), t, new_match_path_detector)
        log_atom_decimal_integer_value_me = LogAtom(self.match_context_decimal_integer_value_me.match_data,
                                                    ParserMatch(self.match_element_decimal_integer_value_me), t, new_match_path_detector)

        self.assertTrue(new_match_path_detector.receive_atom(log_atom_fixed_dme))
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string % (
            datetime.fromtimestamp(t).strftime(self.datetime_format_string), new_match_path_detector.__class__.__name__, description, 1,
            self.match_path_s1, self.pid))
        self.reset_output_stream()

        # repeating should NOT produce the same result
        self.assertTrue(new_match_path_detector.receive_atom(log_atom_fixed_dme))
        self.assertEqual(self.output_stream.getvalue(), '')
        self.reset_output_stream()

        # other MatchElement
        self.assertTrue(new_match_path_detector.receive_atom(log_atom_decimal_integer_value_me))
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string % (
            datetime.fromtimestamp(t).strftime(self.datetime_format_string), new_match_path_detector.__class__.__name__, description, 1,
            self.match_path_d1, self.uid))

    def test3_log_atom_known_from_persisted_data(self):
        """The persisting and reading of permitted log lines should be checked with this test."""
        description = "Test3NewMatchPathDetector"
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True,
                                                       output_logline=False)
        self.analysis_context.register_component(new_match_path_detector, description)
        t = round(time.time(), 3)
        log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_fixed_dme), t, new_match_path_detector)

        self.assertTrue(new_match_path_detector.receive_atom(log_atom_fixed_dme))
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string % (
            datetime.fromtimestamp(t).strftime(self.datetime_format_string), new_match_path_detector.__class__.__name__, description, 1,
            self.match_path_s1, self.pid))
        new_match_path_detector.do_persist()
        self.reset_output_stream()

        otherNewMatchPathDetector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', False,
                                                         output_logline=False)
        otherLogAtomFixedDME = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_fixed_dme), t, otherNewMatchPathDetector)

        self.assertTrue(otherNewMatchPathDetector.receive_atom(otherLogAtomFixedDME))
        self.assertEqual(self.output_stream.getvalue(), '')

    def test4_get_time_trigger_class(self):
        """
        The known paths are to be periodically stored after a certain time. This requires a synchronization class.
        The return of the correct class is to be checked in this test case.
        """
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True,
                                                       output_logline=False)
        self.assertEqual(new_match_path_detector.get_time_trigger_class(), 1)

    # skipcq: PYL-W0105
    """The following test cases should check if the doTimer() method is working properly.This includes the updating of nextPersistTime.
    As it is not updated directly in the method this test cases are not correct. Due to that they are commented."""
    # def test5_do_timer_next_persist_time_none(self):
    #     """During initialization, the next time is not determined (the value is initialized with None). In this case, the persistence is
    #     expected to occur after 600 milliseconds."""
    #     self.new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True,
    #                                                         output_logline=False)
    #     self.assertEqual(self.new_match_path_detector.do_timer(200), 600)
    #     self.assertEqual(self.new_match_path_detector.do_timer(400), 600)
    #     self.assertEqual(self.new_match_path_detector.do_timer(10000), 600)
    #
    # def test6_do_timer_delta_smaller_or_equal_zero(self):
    #     """If the NextPersistTime is less than or equal to zero, the data must be saved."""
    #     self.new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True,
    #                                                         output_logline=False)
    #     self.new_match_path_detector.nextPersistTime = 400
    #     self.assertEqual(self.new_match_path_detector.do_timer(400), 600)
    #     self.assertEqual(self.new_match_path_detector.do_timer(1000), 600)
    #
    # def test7_do_timer_delta_greater_zero(self):
    #     """If the delta does not fall below the limit value, only the delta value should be returned."""
    #     # this test fails due to the missing update of the nextPersistTime variable in the doTimer method
    #     self.new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True,
    #                                                         output_logline=False)
    #     self.new_match_path_detector.nextPersistTime = 400
    #     self.assertEqual(self.new_match_path_detector.do_timer(200), 200)
    #     self.assertEqual(self.new_match_path_detector.do_timer(200), 600)
    #     self.assertEqual(self.new_match_path_detector.do_timer(100), 500)

    def test8_allowlist_event_type_exception(self):
        """This test case checks whether an exception is thrown when entering an event of another class."""
        description = "Test8NewMatchPathDetector"
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True,
                                                       output_logline=False)
        self.analysis_context.register_component(new_match_path_detector, description)
        t = round(time.time(), 3)
        log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_fixed_dme), t, new_match_path_detector)
        new_match_path_detector.receive_atom(log_atom_fixed_dme)
        new_match_path_value_combo_detector = NewMatchPathValueComboDetector(self.aminer_config, [], [self.stream_printer_event_handler],
                                                                             'Default', True, True)
        self.assertRaises(
            Exception, new_match_path_detector.allowlist_event, self.analysis % new_match_path_value_combo_detector.__class__.__name__,
            self.output_stream.getvalue(), None)

    def test9_allowlist_event_allowlisting_data_exception(self):
        """The NewMatchPathDetector can not handle allowlisting data and therefore an exception is expected."""
        description = "Test9NewMatchPathDetector"
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True,
                                                       output_logline=False)
        self.analysis_context.register_component(new_match_path_detector, description)
        t = round(time.time(), 3)
        log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_fixed_dme), t, new_match_path_detector)
        new_match_path_detector.receive_atom(log_atom_fixed_dme)
        self.assertRaises(Exception, new_match_path_detector.allowlist_event, self.analysis % new_match_path_detector.__class__.__name__,
                          self.output_stream.getvalue(), ['random', 'Data'])

    def test10_allowlist_event_with_known_and_unknown_paths(self):
        """This test case checks in which cases an event is triggered and compares with expected results."""
        description = "Test10NewMatchPathDetector"
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True,
                                                       output_logline=False)
        self.analysis_context.register_component(new_match_path_detector, description)
        t = round(time.time(), 3)
        log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_fixed_dme), t, new_match_path_detector)
        new_match_path_detector.receive_atom(log_atom_fixed_dme)
        self.assertEqual(new_match_path_detector.allowlist_event(
            self.analysis % new_match_path_detector.__class__.__name__, self.match_element_fixed_dme.get_path(), None),
            'Allowlisted path(es) %s in %s.' % (
                self.match_element_fixed_dme.get_path(), self.analysis % new_match_path_detector.__class__.__name__))

        new_match_path_detector.learn_mode = False
        self.assertEqual(new_match_path_detector.allowlist_event(
            self.analysis % new_match_path_detector.__class__.__name__, self.match_element_decimal_integer_value_me.get_path(), None),
            'Allowlisted path(es) %s in %s.' % (
                self.match_element_decimal_integer_value_me.path, self.analysis % new_match_path_detector.__class__.__name__))

    # '''
    # This test case checks what happens when no EventHandler is used in the parameters. Requires type check (not yet implemented).
    # '''
    # def test11_fuzzing_anomaly_event_handler(self):
    #   self.new_match_path_detector = NewMatchPathDetector(self.aminer_config, None, 'Default', True, output_logline=False)
    #   t = datetime.fromtimestamp(time.time())
    #   self.log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data,
    #     ParserMatch(self.match_context_fixed_dme), t, self.new_match_path_detector)
    #   self.assertRaises(AttributeError, self.new_match_path_detector.receive_atom, self.log_atom_fixed_dme)
    #
    #   #At least one EventHandler should be used, else the Detector can not report anomalies
    #   self.new_match_path_detector = NewMatchPathDetector(self.aminer_config, [], 'Default', True)
    #   self.log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data,
    #     ParserMatch(self.match_element_fixed_dme), t, self.new_match_path_detector)
    #   self.assertRaises(Exception, self.new_match_path_detector.receive_atom, self.log_atom_fixed_dme)
    #
    # '''
    # An attempt is made to use a non-Boolean expression for the autoIncludeFlag. Requires type check (not yet implemented).
    # '''
    # def test12_fuzzing_learn_mode(self):
    #   self.assertRaises(ArgumentTypeError, NewMatchPathDetector, self.aminer_config,
    #     [self.stream_printer_event_handler], 'Default', None)
    #   self.assertRaises(ArgumentTypeError, NewMatchPathDetector, self.aminer_config,
    #     [self.stream_printer_event_handler], 'Default', 'True')
    #
    # '''
    # An exception is expected if no LogAtom is passed as a parameter. Requires type check (not yet implemented).
    # '''
    # def test13_fuzzing_log_atom(self):
    #   self.new_match_path_detector = NewMatchPathDetector(self.aminer_config,
    #     [self.stream_printer_event_handler], 'Default', True)
    #   self.assertRaises(ArgumentTypeError, self.new_match_path_detector.receive_atom, self.aminer_config)
    #
    # '''
    # The data type must be checked before calculating the remaining time Requires type check (not yet implemented).
    # '''
    # def test14_fuzzing_trigger_time(self):
    #   self.new_match_path_detector = NewMatchPathDetector(self.aminer_config,
    #     [self.stream_printer_event_handler], 'Default', True)
    #   self.new_match_path_detector.nextPersistTime = 400
    #   self.assertRaises(ArgumentTypeError, self.new_match_path_detector.do_timer, '200')


if __name__ == "__main__":
    unittest.main()



import unittest
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.SequenceModelElement import SequenceModelElement
from unit.TestBase import TestBase, DummyMatchContext, DummyFixedDataModelElement


class SequenceModelElementTest(TestBase):
    """Unittests for the SequenceModelElement."""

    id_ = "sequence"
    path = "path"
    children = [DummyFixedDataModelElement("0", b"string0 "), DummyFixedDataModelElement("1", b"string1 "),
                DummyFixedDataModelElement("2", b"string2")]
    match_elements = [MatchElement("path/sequence/0", b"string0 ", b"string0 ", None),
                      MatchElement("path/sequence/1", b"string1 ", b"string1 ", None),
                      MatchElement("path/sequence/2", b"string2", b"string2", None)]

    def test1get_match_element_valid_match(self):
        """Parse matching substring from MatchContext and check if the MatchContext was updated with all characters."""
        data = b"string0 string1 string2"
        match_context = DummyMatchContext(data)
        sequence_me = SequenceModelElement(self.id_, self.children)
        match_element = sequence_me.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, data, data, self.match_elements)

        data = b"string0 string1 string2 other string follows"
        value = b"string0 string1 string2"
        match_context = DummyMatchContext(data)
        sequence_me = SequenceModelElement(self.id_, self.children)
        match_element = sequence_me.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, self.match_elements)

    def test2get_match_element_no_match(self):
        """Parse not matching substring from MatchContext and check if the MatchContext was not changed."""
        data = b""
        match_context = DummyMatchContext(data)
        sequence_me = SequenceModelElement(self.id_, self.children)
        match_element = sequence_me.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"string0 string1 "
        match_context = DummyMatchContext(data)
        sequence_me = SequenceModelElement(self.id_, self.children)
        match_element = sequence_me.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"string0 string1 string3"
        match_context = DummyMatchContext(data)
        sequence_me = SequenceModelElement(self.id_, self.children)
        match_element = sequence_me.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"string0 string0 string2"
        match_context = DummyMatchContext(data)
        sequence_me = SequenceModelElement(self.id_, self.children)
        match_element = sequence_me.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

    def test3element_id_input_validation(self):
        """Check if element_id is validated."""
        self.assertRaises(ValueError, SequenceModelElement, "", self.children)
        self.assertRaises(TypeError, SequenceModelElement, None, self.children)
        self.assertRaises(TypeError, SequenceModelElement, b"path", self.children)
        self.assertRaises(TypeError, SequenceModelElement, True, self.children)
        self.assertRaises(TypeError, SequenceModelElement, 123, self.children)
        self.assertRaises(TypeError, SequenceModelElement, 123.22, self.children)
        self.assertRaises(TypeError, SequenceModelElement, {"id": "path"}, self.children)
        self.assertRaises(TypeError, SequenceModelElement, ["path"], self.children)
        self.assertRaises(TypeError, SequenceModelElement, [], self.children)
        self.assertRaises(TypeError, SequenceModelElement, (), self.children)
        self.assertRaises(TypeError, SequenceModelElement, set(), self.children)

    def test4get_match_element_match_context_input_validation(self):
        """Check if an exception is raised, when other classes than MatchContext are used in get_match_element."""
        model_element = SequenceModelElement(self.id_, self.children)
        data = b"string0 string1 string2"
        model_element.get_match_element(self.path, DummyMatchContext(data))
        model_element.get_match_element(self.path, MatchContext(data))

        self.assertRaises(AttributeError, model_element.get_match_element, self.path, MatchElement(None, data, None, None))
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data.decode())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123.22)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, True)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, None)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, [])
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, {"key": MatchContext(data)})
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, set())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, ())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, model_element)


if __name__ == "__main__":
    unittest.main()



"""
This package contains various classes to build check rulesets.
The ruleset also supports parallel rule evaluation, e.g. the two
rules "A and B and C" and "A and B and D" will only peform the
checks for A and B once, then performs check C and D and trigger
a match action.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import datetime
import sys
import abc
import logging

from aminer.util.History import LogarithmicBackoffHistory
from aminer.util.History import ObjectHistory

from aminer.analysis.AtomFilters import SubhandlerFilter
from aminer.AminerConfig import DEBUG_LOG_NAME, STAT_LOG_NAME
from aminer import AminerConfig

result_string = '%s(%s)'


class MatchAction(metaclass=abc.ABCMeta):
    """This is the interface of all match actions."""

    @abc.abstractmethod
    def match_action(self, log_atom):
        """
        Invoke this method if a rule has matched.
        @param log_atom the LogAtom matching the rules.
        """


class EventGenerationMatchAction(MatchAction):
    """This generic match action forwards information about a rule match on parsed data to a list of event handlers."""

    def __init__(self, event_type, event_message, event_handlers):
        self.event_type = event_type
        self.event_message = event_message
        self.event_handlers = event_handlers

    def match_action(self, log_atom):
        """
        Invoke this method if a rule has matched.
        @param log_atom the LogAtom matching the rules.
        """
        event_data = {}
        for handler in self.event_handlers:
            handler.receive_event(self.event_type, self.event_message, [log_atom.parser_match.match_element.annotate_match('')], event_data,
                                  log_atom, self)


class AtomFilterMatchAction(MatchAction, SubhandlerFilter):
    """This generic match rule forwards all rule matches to a list of AtomHandlerInterface instances using the SubhandlerFilter."""

    def __init__(self, subhandler_list, stop_when_handled_flag=False):
        SubhandlerFilter.__init__(self, subhandler_list, stop_when_handled_flag)

    def match_action(self, log_atom):
        """
        Invoke this method if a rule has matched.
        @param log_atom the LogAtom matching the rules.
        """
        self.receive_atom(log_atom)


class MatchRule(metaclass=abc.ABCMeta):
    """This is the interface of all match rules."""

    log_success = 0
    log_total = 0

    @abc.abstractmethod
    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""

    def log_statistics(self, rule_id):
        """Log statistics of an MatchRule. Override this method for more sophisticated statistics output of the MatchRule."""
        if AminerConfig.STAT_LEVEL > 0:
            logging.getLogger(STAT_LOG_NAME).info(f"Rule '{rule_id}' processed {self.log_success} out of {self.log_total} log atoms"
                                                  f" successfully in the last 60 minutes.")
        self.log_success = 0
        self.log_total = 0
        if hasattr(self, 'sub_rules'):
            for i, rule in enumerate(self.sub_rules):
                rule.log_statistics(rule_id + '.' + rule.__class__.__name__ + str(i))
        if hasattr(self, 'rule_lookup_dict'):
            for i, rule_key in enumerate(self.rule_lookup_dict):
                rule = self.rule_lookup_dict[rule_key]
                rule.log_statistics(rule_id + '.' + rule.__class__.__name__ + str(i))
        if hasattr(self, 'default_rule'):
            self.default_rule.log_statistics(rule_id + '.default_rule.' + self.default_rule.__class__.__name__)
        if hasattr(self, 'sub_rule'):
            self.sub_rule.log_statistics(rule_id + '.' + self.sub_rule.__class__.__name__)


class AndMatchRule(MatchRule):
    """This class provides a rule to match all subRules (logical and)."""

    def __init__(self, sub_rules, match_action=None):
        """
        Create the rule.
        @param match_action if None, no action is performed.
        """
        self.sub_rules = sub_rules
        self.match_action = match_action

    def match(self, log_atom):
        """
        Check if this rule matches. Rule evaluation will stop when the first match fails.
        If a matchAction is attached to this rule, it will be invoked at the end of all checks.
        @return True when all subrules matched.
        """
        self.log_total += 1
        for rule in self.sub_rules:
            if not rule.match(log_atom):
                return False
        if self.match_action is not None:
            self.match_action.match_action(log_atom)
        self.log_success += 1
        return True

    def __str__(self):
        result = ''
        preamble = ''
        for match_element in self.sub_rules:
            result += result_string % (preamble, match_element)
            preamble = ' and '
        return result


class OrMatchRule(MatchRule):
    """This class provides a rule to match any subRules (logical or)."""

    def __init__(self, sub_rules, match_action=None):
        """
        Create the rule.
        @param match_action if None, no action is performed.
        """
        self.sub_rules = sub_rules
        self.match_action = match_action

    def match(self, log_atom):
        """
        Check if this rule matches. Rule evaluation will stop when the first match succeeds.
        If a matchAction is attached to this rule, it will be invoked after the first match.
        @return True when any subrule matched.
        """
        self.log_total += 1
        for rule in self.sub_rules:
            if rule.match(log_atom):
                if self.match_action is not None:
                    self.match_action.match_action(log_atom)
                self.log_success += 1
                return True
        return False

    def __str__(self):
        result = ''
        preamble = ''
        for match_element in self.sub_rules:
            result += result_string % (preamble, match_element)
            preamble = ' or '
        return result


class ParallelMatchRule(MatchRule):
    """
    This class is a rule testing all the subrules in parallel.
    From the behaviour it is similar to the OrMatchRule, returning true if any subrule matches. The difference is that matching will not
    stop after the first positive match. This does only make sense when all subrules have match actions associated.
    """

    def __init__(self, sub_rules, match_action=None):
        """
        Create the rule.
        @param match_action if None, no action is performed.
        """
        self.sub_rules = sub_rules
        self.match_action = match_action

    def match(self, log_atom):
        """
        Check if any of the subrules rule matches. The matching procedure will not stop after the first positive match.
        If a matchAction is attached to this rule, it will be invoked at the end of all checks.
        @return True when any subrule matched.
        """
        self.log_total += 1
        match_flag = False
        for rule in self.sub_rules:
            if rule.match(log_atom):
                match_flag = True
        if match_flag and (self.match_action is not None):
            self.match_action.match_action(log_atom)
        if match_flag:
            self.log_success += 1
        return match_flag

    def __str__(self):
        result = ''
        preamble = ''
        for match_element in self.sub_rules:
            result += result_string % (preamble, match_element)
            preamble = ' por '
        return result


class ValueDependentDelegatedMatchRule(MatchRule):
    """
    This class is a rule delegating rule checking to subrules depending on values found within the parser_match.
    The result of this rule is the result of the selected delegation rule.
    """

    def __init__(self, target_path_list, rule_lookup_dict, default_rule=None, match_action=None):
        """
        Create the rule.
        @param list with value paths that are used to extract the lookup keys for ruleLookupDict. If value lookup fails, None
        will be used for lookup.
        @param rule_lookup_dict dictionary with tuple containing values for valuePathList as key and target rule as value.
        @param default_rule when not none, this rule will be executed as default. Otherwise, when rule lookup failed, False will
        be returned unconditionally.
        @param match_action if None, no action is performed.
        """
        self.target_path_list = target_path_list
        self.rule_lookup_dict = rule_lookup_dict
        self.default_rule = default_rule
        self.match_action = match_action

    def match(self, log_atom):
        """
        Try to locate a rule for delegation or use the default rule.
        @return True when selected delegation rule matched.
        """
        self.log_total += 1
        match_dict = log_atom.parser_match.get_match_dictionary()
        value_list = []
        for path in self.target_path_list:
            value_element = match_dict.get(path)
            if value_element is not None:
                value_list.append(value_element.match_object)
        if len(value_list) > 0:
            value = tuple(value_list)
        else:
            value = None
        rule = self.rule_lookup_dict.get(value, self.default_rule)
        if rule is None:
            return False
        if rule.match(log_atom):
            if self.match_action is not None:
                self.match_action.match_action(log_atom)
            self.log_success += 1
            return True
        return False

    def __str__(self):
        result = 'ValueDependentDelegatedMatchRule'
        return result


class NegationMatchRule(MatchRule):
    """Match elements of this class return true when the subrule did not match."""

    def __init__(self, sub_rule, match_action=None):
        self.sub_rule = sub_rule
        self.match_action = match_action

    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""
        self.log_total += 1
        if self.sub_rule.match(log_atom):
            return False
        if self.match_action is not None:
            self.match_action.match_action(log_atom)
        self.log_success += 1
        return True

    def __str__(self):
        return f'not {self.sub_rule}'


class PathExistsMatchRule(MatchRule):
    """Match elements of this class return true when the given target_path was found in the parsed match data."""

    def __init__(self, target_path, match_action=None):
        self.target_path = target_path
        self.match_action = match_action

    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""
        self.log_total += 1
        if self.target_path in log_atom.parser_match.get_match_dictionary():
            if self.match_action is not None:
                self.match_action.match_action(log_atom)
            self.log_success += 1
            return True
        return False

    def __str__(self):
        return f'hasPath({self.target_path})'


class ValueMatchRule(MatchRule):
    """Match elements of this class return true when the given target_path exists and has exactly the given parsed value."""

    def __init__(self, target_path, value, match_action=None):
        self.target_path = target_path
        self.value = value
        self.match_action = match_action

    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""
        self.log_total += 1
        test_value = log_atom.parser_match.get_match_dictionary().get(self.target_path, None)
        if test_value is not None:
            if isinstance(self.value, bytes) and isinstance(test_value.match_object, str) and test_value.match_object is not None:
                test_value.match_object = test_value.match_object.encode()
            elif isinstance(self.value, str) and isinstance(test_value.match_object, bytes) and self.value is not None:
                self.value = self.value.encode()
            elif not isinstance(self.value, type(test_value.match_object)):
                raise TypeError(f"The type of the value of the ValueMatchRule does not match the test_value. value: {type(self.value)}, "
                                f"test_value: {type(test_value.match_object)}")
        if (test_value is not None) and (test_value.match_object == self.value):
            if self.match_action is not None:
                self.match_action.match_action(log_atom)
            self.log_success += 1
            return True
        return False

    def __str__(self):
        if isinstance(self.value, bytes):
            self.value = self.value.decode()
        return f'value({self.target_path})=={self.value}'


class ValueListMatchRule(MatchRule):
    """Match elements of this class return true when the given path exists and has exactly one of the values included in the value list."""

    def __init__(self, target_path, target_value_list, match_action=None):
        self.target_path = target_path
        self.target_value_list = target_value_list
        self.match_action = match_action

    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""
        self.log_total += 1
        test_value = log_atom.parser_match.get_match_dictionary().get(self.target_path)
        if (test_value is not None) and (test_value.match_object in self.target_value_list):
            if self.match_action is not None:
                self.match_action.match_action(log_atom)
            self.log_success += 1
            return True
        return False

    def __str__(self):
        return f"value({' '.join([str(value) for value in self.target_value_list])}) in {self.target_path}"


class ValueRangeMatchRule(MatchRule):
    """Match elements of this class return true when the given target_path exists and the value is included in [lower, upper] range."""

    def __init__(self, target_path, lower_limit, upper_limit, match_action=None):
        self.target_path = target_path
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.match_action = match_action

    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""
        self.log_total += 1
        test_value = log_atom.parser_match.get_match_dictionary().get(self.target_path, None)
        if test_value is None:
            return False
        test_value = test_value.match_object
        if self.lower_limit <= test_value <= self.upper_limit:
            if self.match_action is not None:
                self.match_action.match_action(log_atom)
            self.log_success += 1
            return True
        return False

    def __str__(self):
        return f'value({self.target_path}) inrange ({self.lower_limit}, {self.upper_limit})'


class StringRegexMatchRule(MatchRule):
    """Elements of this class return true when the given path exists and the string repr of the value matches the regular expression."""

    def __init__(self, target_path, match_regex, match_action=None):
        self.target_path = target_path
        self.match_regex = match_regex
        self.match_action = match_action

    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""
        self. log_total += 1
        # Use the class object as marker for nonexisting entries
        test_value = log_atom.parser_match.get_match_dictionary().get(self.target_path, None)
        if (test_value is None) or (self.match_regex.match(test_value.match_string) is None):
            return False
        if self.match_action is not None:
            self.match_action.match_action(log_atom)
        self.log_success += 1
        return True

    def __str__(self):
        return f'string({self.target_path}) =regex= {self.match_regex.pattern}'


class ModuloTimeMatchRule(MatchRule):
    """
    Match elements of this class return true when the following conditions are met.
    The given target_path exists, denotes a datetime object and the seconds since 1970 from that date modulo the given value are included in
    [lower, upper] range.
    """

    def __init__(self, target_path, seconds_modulo, lower_limit, upper_limit, match_action=None, tzinfo=None):
        """
        @param target_path the target_path to the datetime object to use to evaluate the modulo time rules on.
        When None, the default timestamp associated with the match is used.
        """
        self.target_path = target_path
        self.seconds_modulo = seconds_modulo
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.match_action = match_action
        self.tzinfo = tzinfo
        if tzinfo is None:
            self.tzinfo = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""
        self.log_total += 1
        test_value = None
        if self.target_path is None:
            test_value = log_atom.get_timestamp()
        else:
            time_match = log_atom.parser_match.get_match_dictionary().get(self.target_path, None)
            if time_match is None:
                return False
            test_value = time_match.match_object + datetime.datetime.now(self.tzinfo).utcoffset().total_seconds()

        if test_value is None:
            return False
        test_value %= self.seconds_modulo
        if self.lower_limit <= test_value <= self.upper_limit:
            if self.match_action is not None:
                self.match_action.match_action(log_atom)
            self.log_success += 1
            return True
        return False


class ValueDependentModuloTimeMatchRule(MatchRule):
    """
    Match elements of this class return true when the following conditions are met.
    The given path exists, denotes a datetime object and the seconds since 1970 rom that date modulo the given value are included in a
    [lower, upper] range selected by values from the match.
    """

    def __init__(
            self, target_path, seconds_modulo, target_path_list, limit_lookup_dict, default_limit=None, match_action=None, tzinfo=None):
        """
        @param target_path the target_path to the datetime object to use to evaluate the modulo time rules on.
        When None, the default timestamp associated with the match is used.
        @param default_limit use this default limit when limit lookup failed. Without a default limit, a failed lookup will cause
        the rule not to match.
        """
        self.target_path = target_path
        self.seconds_modulo = seconds_modulo
        self.target_path_list = target_path_list
        self.limit_lookup_dict = limit_lookup_dict
        self.default_limit = default_limit
        self.match_action = match_action
        self.tzinfo = tzinfo
        if tzinfo is None:
            self.tzinfo = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""
        self.log_total += 1
        match_dict = log_atom.parser_match.get_match_dictionary()
        value_list = []
        for path in self.target_path_list:
            value_element = match_dict.get(path)
            if value_element is not None:
                value_list.append(value_element.match_object)

        if len(value_list) > 0:
            value = value_list[0]
        else:
            value = None
        limits = self.limit_lookup_dict.get(value, self.default_limit)
        if limits is None:
            return False

        test_value = None
        if self.target_path is None:
            test_value = log_atom.get_timestamp()
        else:
            time_match = log_atom.parser_match.get_match_dictionary().get(self.target_path, None)
            if time_match is None:
                return False
            test_value = time_match.match_object + datetime.datetime.now(self.tzinfo).utcoffset().total_seconds()

        if test_value is None:
            return False
        test_value %= self.seconds_modulo
        if limits[0] <= test_value <= limits[1]:
            if self.match_action is not None:
                self.match_action.match_action(log_atom)
            self.log_success += 1
            return True
        return False


class IPv4InRFC1918MatchRule(MatchRule):
    """
    Match elements of this class return true when the path matches and contains a valid IPv4 address from the RFC1918 private IP ranges.
    This could also be done by distinct range match elements, but as this kind of matching is common, have an own element for it.
    """

    def __init__(self, target_path, match_action=None):
        self.target_path = target_path
        self.match_action = match_action

    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""
        self.log_total += 1
        match_element = log_atom.parser_match.get_match_dictionary().get(self.target_path)
        if (match_element is None) or not isinstance(match_element.match_object, int):
            return False
        value = match_element.match_object
        if ((value & 0xff000000) == 0xa000000) or ((value & 0xfff00000) == 0xac100000) or ((value & 0xffff0000) == 0xc0a80000):
            if self.match_action is not None:
                self.match_action.match_action(log_atom)
            self.log_success += 1
            return True
        return False

    def __str__(self):
        return f'hasPath({self.target_path})'


class DebugMatchRule(MatchRule):
    """
    This rule can be inserted into a normal ruleset just to see when a match attempt is made.
    It just prints out the current log_atom that is evaluated. The match action is always invoked when defined, no matter which match
    result is returned.
    """

    def __init__(self, debug_match_result=False, match_action=None):
        self.debug_match_result = debug_match_result
        self.match_action = match_action

    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""
        self.log_total += 1
        print(f'Rules.DebugMatchRule: triggered while handling "{repr(log_atom.parser_match.match_element.match_string)}"',
              file=sys.stderr)
        if self.match_action is not None:
            self.match_action.match_action(log_atom)
        self.log_success += 1
        return self.debug_match_result

    def __str__(self):
        return f'{self.debug_match_result}'


class DebugHistoryMatchRule(MatchRule):
    """
    This rule can be inserted into a normal ruleset just to see when a match attempt is made.
    It just adds the evaluated log_atom to a ObjectHistory.
    """

    def __init__(self, object_history=None, debug_match_result=False, match_action=None):
        """
        Create a DebugHistoryMatchRule object.
        @param object_history use this ObjectHistory to collect the LogAtoms. When None, a default LogarithmicBackoffHistory for 10 items.
        """
        if object_history is None:
            object_history = LogarithmicBackoffHistory(10)
        elif not isinstance(object_history, ObjectHistory):
            msg = 'object_history is not an instance of ObjectHistory'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        self.object_history = object_history
        self.debug_match_result = debug_match_result
        self.match_action = match_action

    def match(self, log_atom):
        """Check if this rule matches. On match an optional match_action could be triggered."""
        self.log_total += 1
        self.object_history.add_object(log_atom)
        if self.match_action is not None:
            self.match_action.match_action(log_atom)
        self.log_success += 1
        return self.debug_match_result

    def get_history(self):
        """Get the history object from this debug rule."""
        return self.object_history



"""
This module defines a detector for unsorted timestamps.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import os

from aminer.input.InputInterfaces import AtomHandlerInterface
from datetime import datetime
from aminer.AminerConfig import CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX
from aminer import AminerConfig


class TimestampsUnsortedDetector(AtomHandlerInterface):
    """
    This class creates events when unsorted timestamps are detected.
    This is useful mostly to detect algorithm malfunction or configuration errors, e.g. invalid timezone configuration.
    """

    def __init__(self, aminer_config, anomaly_event_handlers, exit_on_error_flag=False, output_logline=True):
        """
        Initialize the detector.
        @param aminer_config configuration from analysis_context.
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        @param exit_on_error_flag exit the aminer forcefully if a log atom with a wrong timestamp is found.
        @param output_logline specifies whether the full parsed log atom should be provided in the output.
        """
        # avoid "defined outside init" issue
        self.log_success, self.log_total = [None]*2
        super().__init__(
            aminer_config=aminer_config, anomaly_event_handlers=anomaly_event_handlers, exit_on_error_flag=exit_on_error_flag,
            output_logline=output_logline
        )
        self.last_timestamp = 0

    def receive_atom(self, log_atom):
        """
        Receive on parsed atom and the information about the parser match.
        @param log_atom the parsed log atom
        @return True if this handler was really able to handle and process the match. Depending on this information, the caller
        may decide if it makes sense passing the parsed atom also to other handlers.
        """
        self.log_total += 1
        if log_atom.get_timestamp() is None:
            return False
        if log_atom.get_timestamp() < self.last_timestamp:
            try:
                data = log_atom.raw_data.decode(AminerConfig.ENCODING)
            except UnicodeError:
                data = repr(log_atom.raw_data)
            original_log_line_prefix = self.aminer_config.config_properties.get(CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
            if self.output_logline:
                sorted_log_lines = [log_atom.parser_match.match_element.annotate_match('') + os.linesep + original_log_line_prefix + data]
            else:
                sorted_log_lines = [original_log_line_prefix + data]
            analysis_component = {'LastTimestamp': self.last_timestamp}
            event_data = {'AnalysisComponent': analysis_component}
            for listener in self.anomaly_event_handlers:
                listener.receive_event(
                    f'Analysis.{self.__class__.__name__}',
                    f"Timestamp {datetime.fromtimestamp(log_atom.get_timestamp()).strftime('%Y-%m-%d %H:%M:%S')} below "
                    f"{datetime.fromtimestamp(self.last_timestamp).strftime('%Y-%m-%d %H:%M:%S')}", sorted_log_lines, event_data,
                    log_atom, self)
            if self.exit_on_error_flag:
                import sys
                sys.exit(1)
        self.last_timestamp = log_atom.get_timestamp()
        self.log_success += 1
        return True
